/*
 * Segment Datenübernahme und Aufbereitung (DUA), Fehleranalyse TLS spontan
 * Copyright (C) 2016 by Regierungspräsidium Tübingen, Landesstelle für Straßentechnik, Thomas Bräuner
 * Basiert auf DeFaApplikation aus de.bsvrz.dua.fehlertls.fehlertls, BitCtrl Systems GmbH
 * 
 * SWE de.bsvrz.dua.fehlertlsspontan (Fehleranalyse TLS spontan)
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the 
 * GNU General Public License as published by the Free Software Foundation; either version 3 of 
 * the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 *
 * Dieses Programm ist freie Software. Sie können es unter den Bedingungen der GNU General Public 
 * License, wie von der Free Software Foundation veröffentlicht, weitergeben und/oder modifizieren, 
 * entweder gemäß Version 3 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
 * Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, dass es Ihnen von Nutzen sein wird, 
 * aber OHNE IRGENDEINE GARANTIE, sogar ohne die implizite Garantie der MARKTREIFE oder der 
 * VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License.
 * Sie sollten ein Exemplar der GNU General Public License zusammen mit diesem Programm erhalten haben. 
 * Falls nicht, siehe <http://www.gnu.org/licenses/>.
 * Eine deutsche Übersetzung findet sich unter http://www.gnu.de/documents/gpl.de.html.
 *
 * Contact Information/Kontakt:
 * Regierungspräsidium Tübingen
 * Landesstelle für Straßentechnik
 * Heilbronner Straße 300 - 302
 * 70469 Stuttgart
 * Phone: +49 0711 8910-200
 * mailto: abteilung9@rpt.bwl.de
 */

package de.bsvrz.dua.fehlertlsspontan;

import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dua.fehlertlsspontan.parameter.ParameterTlsFehlerAnalyse;
import de.bsvrz.dua.fehlertlsspontan.tls.TlsHierarchie;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Diese SWE dient zur Ermittlung der Fehlerursache bei DE (Datenendgerät gemäß
 * TLS), die Daten nicht zyklisch sondern nur spontan senden (WZG, VLT). Es ergänzt
 * somit die SWE "Fehleranalyse fehlende Messdaten TLS" um die dort nicht 
 * betrachteten DE-Typen.
 * Da es bei spontan sendenden DE keine Überwachung eines erwarteten Messwerts
 * geben kann, überwacht diese SWE nur den Fehlerstatus dieser DE-Typen. Aber auch
 * hier wird die gesamte Kommunikationskette zwischen Detektor, EAK, SM, KRI, UZ 
 * und VRZ betrachtet und somit versucht, die Quelle der Fehlerursache zu
 * lokalisieren. Das Ergebnis der Analyse wird ebenfalls über Betriebsmeldungen
 * und Fehlerstatusausgaben pro DE verfügbar gemacht.
 * 
 * @author Regierungspräsidium Tübingen, Landesstelle für Straßentechnik, Thomas Bräuner
 * @author BitCtrl Systems GmbH, Thierfelder
 * 
 * @version 1.0.0
 */
public class FehlerTlsSpontan implements StandardApplication {

	/**
	 * das Systemobjekt vom Typ <code>typ.tlsFehlerAnalyse</code>, mit dem diese
	 * Applikation assoziiert ist (aus der sie ihre Parameter bezieht).
	 */
	private static SystemObject tlsFehlerAnalyseObjekte = null;
	
	/**
	 * die PIDs der Geraete, die in der Kommandozeile uebergeben wurden.
	 */
	private String[] geraetePids = null;

	/**
	 * Geräte, die in der Kommandozeile uebergeben wurden.
	 */
	private Set<SystemObject> geraete = new HashSet<>();

	/**
	 * die Pid des Objektes vom Typ <code>typ.tlsFehlerAnalyse</code> mit dem
	 * diese Applikation assoziiert ist (aus der sie ihre Parameter bezieht).
	 */
	private String parameterModulPid = null;

	/**
	 * Erfragt das Systemobjekt vom Typ <code>typ.tlsFehlerAnalyse</code>, mit
	 * dem diese Applikation assoziiert ist (aus der sie ihre Parameter
	 * bezieht).
	 * 
	 * @return das Systemobjekt vom Typ <code>typ.tlsFehlerAnalyse</code>, mit
	 *         dem diese Applikation assoziiert ist (aus der sie ihre Parameter
	 *         bezieht)
	 */
	public static final SystemObject getTlsFehlerAnalyseObjekt() {
		return tlsFehlerAnalyseObjekte;
	}

	@Override
	public void parseArguments(ArgumentList argumente) throws Exception {

		Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
			@Override
			public void uncaughtException(Thread t, Throwable e) {
				Debug.getLogger().error("Applikation wird wegen unerwartetem Fehler beendet", e);
				e.printStackTrace();
				Runtime.getRuntime().exit(-1);
			}
		});

		if (argumente.hasArgument("-param")) {
			this.parameterModulPid = argumente.fetchArgument("-param")
					.asNonEmptyString();
		} else {
			Debug.getLogger()
					.warning(
							"Kein Objekt vom Typ \"typ.tlsFehlerAnalyse\" zur Parametrierung dieser Instanz uebergeben (-param=...)");
		}

		this.geraetePids = argumente.fetchArgument("-geraet").asNonEmptyString().split(",");
		argumente.fetchUnusedArguments();
	}

	@Override
	public void initialize(ClientDavInterface dav) throws Exception {
		MessageSender.getInstance().setApplicationLabel("Fehleranalyse TLS spontan");

		for (String pidVonGeraet : this.geraetePids) {
			SystemObject geraeteObjekt = dav.getDataModel().getObject(pidVonGeraet);
			if (geraeteObjekt != null) {
				if (geraeteObjekt.isOfType("typ.gerät")) { 
					this.geraete.add(geraeteObjekt);
				} else {
					Debug.getLogger().warning("Das übergebene Objekt " + pidVonGeraet + " ist nicht vom Typ Gerät.");
				}
			} else {
				Debug.getLogger().warning("Das übergebene Gerät " + pidVonGeraet + " existiert nicht.");
			}
		}

		if (this.parameterModulPid == null) {
			for (SystemObject obj : dav.getDataModel()
					.getType("typ.tlsFehlerAnalyse").getElements()) { //$NON-NLS-1$
				if (obj.isValid()) {
					if (tlsFehlerAnalyseObjekte != null) {
						Debug.getLogger()
								.warning(
										"Es existieren mehrere Objekte vom Typ \"typ.tlsFehlerAnalyse\""); //$NON-NLS-1$
						break;
					}
					tlsFehlerAnalyseObjekte = obj;
					if (obj.getConfigurationArea().equals(
							dav.getDataModel().getConfigurationAuthority()
									.getConfigurationArea())) {
						break;
					}
				}
			}
		} else {
			SystemObject dummy = dav.getDataModel().getObject(
					this.parameterModulPid);
			if (dummy != null && dummy.isValid()) {
				tlsFehlerAnalyseObjekte = dummy;
			}
		}

		if (tlsFehlerAnalyseObjekte == null) {
			throw new RuntimeException(
					"Es existiert kein Objekt vom Typ \"typ.tlsFehlerAnalyse\""); //$NON-NLS-1$
		} else {
			ParameterTlsFehlerAnalyse.getInstanz(dav, tlsFehlerAnalyseObjekte);
			Debug.getLogger().config(
					"Es werden die Parameter von " + tlsFehlerAnalyseObjekte //$NON-NLS-1$
							+ " verwendet"); //$NON-NLS-1$
		}

		if (this.geraete.isEmpty()) {
			Debug.getLogger().warning("Es wurden keine gültigen Geräte übergeben.");
		} else {
			TlsHierarchie.getInstance(dav, geraete);
		}
	}

	/**
	 * Startet diese Applikation.
	 * 
	 * @param argumente
	 *            Argumente der Kommandozeile
	 */
	public static void main(String[] argumente) {
		StandardApplicationRunner.run(new FehlerTlsSpontan(), argumente);	
	}				
}
