/*
 * Segment Datenübernahme und Aufbereitung (DUA), Fehleranalyse TLS spontan
 * Copyright (C) 2016 by Regierungspräsidium Tübingen, Landesstelle für Straßentechnik
 * Basiert auf AbstraktGeraet aus de.bsvrz.dua.fehlertls.tls, BitCtrl Systems GmbH
 * 
 * SWE de.bsvrz.dua.fehlertlsspontan (Fehleranalyse TLS spontan)
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the 
 * GNU General Public License as published by the Free Software Foundation; either version 3 of 
 * the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 *
 * Dieses Programm ist freie Software. Sie können es unter den Bedingungen der GNU General Public 
 * License, wie von der Free Software Foundation veröffentlicht, weitergeben und/oder modifizieren, 
 * entweder gemäß Version 3 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
 * Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, dass es Ihnen von Nutzen sein wird, 
 * aber OHNE IRGENDEINE GARANTIE, sogar ohne die implizite Garantie der MARKTREIFE oder der 
 * VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License.
 * Sie sollten ein Exemplar der GNU General Public License zusammen mit diesem Programm erhalten haben. 
 * Falls nicht, siehe <http://www.gnu.org/licenses/>.
 * Eine deutsche Übersetzung findet sich unter http://www.gnu.de/documents/gpl.de.html.
 *
 * Contact Information/Kontakt:
 * Regierungspräsidium Tübingen
 * Landesstelle für Straßentechnik
 * Heilbronner Straße 300 - 302
 * 70469 Stuttgart
 * Phone: +49 0711 8910-200
 * mailto: abteilung9@rpt.bwl.de
 */

package de.bsvrz.dua.fehlertlsspontan.tls;

import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dua.fehlertlsspontan.parameter.IParameterTlsFehlerAnalyseListener;
import de.bsvrz.sys.funclib.operatingMessage.OperatingMessage;

/**
 * Abstrakte Repräsentation einer Objektes vom Typ <code>typ.gerät</code>.
 * 
 * @author Regierungspräsidium Tübingen, Landesstelle für Straßentechnik, Thomas Bräuner
 * @author BitCtrl Systems GmbH, Thierfelder
 */
public abstract class TlsGeraet implements IParameterTlsFehlerAnalyseListener {

	/**
	 * Der zusätzliche Zeitverzug, der nach dem erwarteten Empfangszeitpunkt
	 * noch bis zur Erkennung eines nicht gelieferten Messwertes abgewartet
	 * werden muss.
	 */
	private long zeitVerzugFehlerErkennung = Long.MIN_VALUE;

	/**
	 * Der zusätzliche Zeitverzug, der nach der Fehlererkennung bis zur
	 * Fehlerermittlung abgewartet werden muss.
	 */
	private long zeitVerzugFehlerErmittlung = Long.MIN_VALUE;

	/**
	 * ob die TLS-Hierarchie in einem Baum dargestellt werden soll.
	 */
	private static final boolean TLS_BAUM = true;

//	/**
//	 * Zuletzt versendeter Meldungstext
//	 */
//	private String letzteMeldung;

	/**
	 * Mögliche Gerätearten, die fehlerhaft sein können.
	 */
	public enum Art {
		/**
		 * KRI.
		 */
		KRI,

		/**
		 * INSELBUS.
		 */
		INSELBUS,

		/**
		 * SM.
		 */
		SM,

		/**
		 * EAK.
		 */
		EAK,

		/**
		 * DE.
		 */
		DE
	}
	/**
	 * statische Datenverteiler-Verbindung.
	 */
	protected static ClientDavInterface davVerbindung = null;

	/**
	 * Konfigurationsobjekt vom Typ <code>typ.gerät</code>.
	 */
	protected ConfigurationObject objekt = null;

	/**
	 * Die in der TLS-Hierarchie unter diesem Gerät liegenden Geräte.
	 */
	protected Set<TlsGeraet> kinder = new HashSet<>();

	/**
	 * Das in der TLS-Hierarchie über diesem Gerät liegende Gerät.
	 */
	protected TlsGeraet elter = null;

	/**
	 * alle DEs, die sich unterhalb von diesem Element befinden.
	 */
	private Set<De> des = null;

	/**
	 * Erfragt die Geräteart dieses Gerätes.
	 * 
	 * @return die Geräteart dieses Gerätes
	 */
	public abstract Art getGeraeteArt();

	/**
	 * Publiziert einen Fehler.
	 * 
	 * @param zeitStempel
	 *            der Zeitstempel des Fehlers
	 */
	public abstract void publiziereFehler(final long zeitStempel);

	/**
	 * Standardkonstruktor.
	 * 
	 * @param dav
	 *            Datenverteiler-Verbindung
	 * @param objekt
	 *            ein Systemobjekt vom Typ <code>typ.gerät</code>
	 * @param elter
	 *            das in der TLS-Hierarchie über diesem Gerät liegende Gerät
	 */
	protected TlsGeraet(ClientDavInterface dav, SystemObject objekt, TlsGeraet elter) {
		if (davVerbindung == null) {
			davVerbindung = dav;
		}
		this.objekt = (ConfigurationObject) objekt;
		this.elter = elter;
	}

	/**
	 * Erfragt die in der TLS-Hierarchie unter diesem Gerät liegenden Geräte.
	 * 
	 * @return die in der TLS-Hierarchie unter diesem Geraet liegenden Geräte
	 *         (ggf. leere Liste)
	 */
	public final Set<TlsGeraet> getKinder() {
		return this.kinder;
	}

	/**
	 * Erfragt das in der TLS-Hierarchie über diesem Geraet liegende Gerät.
	 * 
	 * @return das in der TLS-Hierarchie über diesem Gerät liegende Gerät
	 *         bzw. <code>null</code>, wenn dieses Gerät die Spitze der
	 *         Hierarchie sein sollte.
	 */
	public final TlsGeraet getVater() {
		return this.elter;
	}

	/**
	 * Erfragt ob dieses Gerät an der Spitze einer TLS-Hierarchie steht.
	 * 
	 * @return ob dieses Gerät an der Spitze einer TLS-Hierarchie steht
	 */
	public final boolean isOberstesElement() {
		return this.elter == null || this.elter.getObjekt() == null;
	}

	/**
	 * {@inheritDoc}
	 * Der Vergleich wird auf Basis der ID des Systemobjekts durchgeführt.
	 */
	@Override
	public boolean equals(Object obj) {
		boolean ergebnis = false;

		if (obj != null && obj instanceof TlsGeraet) {
			TlsGeraet that = (TlsGeraet) obj;
			ergebnis = this.objekt.getId() == that.objekt.getId();
		}

		return ergebnis;
	}

	/**
	 * {@inheritDoc}
	 * Stellt die Einbettung des Objekts in die TLS-Hierarchie dar.
	 */
	@Override
	public String toString() {
		if (TLS_BAUM) {
			String baum = "";

			TlsGeraet dummy = this;
			while (dummy.getVater() != null) {
				baum += "   ";
				dummy = dummy.getVater();
			}
			if (this.objekt == null)
				baum += "WURZEL";
			else if (this.getGeraeteArt() == Art.INSELBUS)
				baum += "Inselbus " + ((Inselbus)this).getPort();
			else
				baum += this.objekt.getPid();
//			baum += this.objekt == null ? "WURZEL" : this.objekt.getPid();
			for (TlsGeraet kind : this.kinder) {
				baum += "\n" + kind.toString();
			}

			return baum;
		} else {
			String e = "Keiner";
			String k = "Keine";

			if (this.elter != null) {
				e = this.elter.getObjekt() == null ? "WURZEL" : this.elter.getObjekt().getPid(); 
			}
			if (!this.kinder.isEmpty()) {
				TlsGeraet[] dummy = this.kinder
						.toArray(new TlsGeraet[0]);
				k = dummy[0].getObjekt().getPid();
				for (int i = 1; i < dummy.length; i++) {
					k += ", " + dummy[i].getObjekt().getPid(); 
				}
			}

			return this.objekt == null ? "WURZEL" : this.objekt.toString() + " (Elter: " + e + ", Kinder:[" + k + "])";			
		}
	}

	/**
	 * Erfragt die in der untersten TLS-Hierarchie unter diesem Gerät liegenden
	 * Geräte (DEs).
	 * 
	 * @return die in der untersten TLS-Hierarchie unter diesem Gerät liegenden
	 *         Geräte (ggf. leere Liste)
	 */
	public final Set<De> getDes() {
		if (this.des == null) {
			synchronized (this) {
				this.des = new HashSet<>();
				sammleDes(this.des);
			}
		}

		return this.des;
	}

	/**
	 * Erfragt das mit diesem Objekt assoziierte Systemobjekt.
	 * 
	 * @return das mit diesem Objekt assoziierte Systemobjekt
	 */
	public final SystemObject getObjekt() {
		return this.objekt;
	}

	/**
	 * Sammelt rekursiv alle DE unterhalb dieses Objektes.
	 * 
	 * @param des1
	 *            eine Menge mit Des
	 */
	private void sammleDes(Set<De> des1) {
		if (this.getGeraeteArt() == Art.DE) {
			des1.add((De) this);
		} else {
			for (TlsGeraet kind : this.kinder) {
				kind.sammleDes(des1);
			}
		}
	}

	/**
	 * Veröffentlicht eine Betriebsmeldung (einmalig)
	 * 
	 * @param message Betriebsmeldung
	 */
	public void publiziere(OperatingMessage message) {
//		if(message.getMessage().equals(letzteMeldung)) {
//			return;
//		}
//		letzteMeldung = message.getMessage();
		message.send();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public synchronized void aktualisiereParameterTlsFehlerAnalyse(
			long zeitverzugFehlerErkennung, long zeitverzugFehlerErmittlung) {
		this.zeitVerzugFehlerErkennung = zeitverzugFehlerErkennung;
		this.zeitVerzugFehlerErmittlung = zeitverzugFehlerErmittlung;
	}
	
	/**
	 * Bestimmt, ob eine Betriebsmeldung abgesetzt werden darf,
	 * weil beide relevanten Parameter ungleich null sind.
	 * 
	 * @return ob eine Betriebsmeldung abgesetzt werden darf
	 */
	public boolean darfBetriebsmeldungSenden() {
		if (zeitVerzugFehlerErkennung > 0 && zeitVerzugFehlerErmittlung > 0)
			return true;
		return false;
	}
	
}
